// Common file used by the different target platforms

using System;
using System.Collections.Generic;
using System.Windows.Forms;
using System.Drawing;
using System.IO;
using Microsoft.Win32;

namespace List
{
    public partial class List
    {
        public Node rootNode;
        public Node visibleNode;

        public string fileName;
        public string _title;

        public string title
        {
            get
            {
                return _title;
            }

            set
            {
                _title = value;
                Text = value;

                UpdateTitleBox();
            }
        }

        public bool modified = false;
        private bool updateInProgress = false;
        private bool inhibitAutoSave = false;

        public Color uncheckedColor = Color.Black;
        public Color checkedColor = Color.Gray;

        public List<String> tagFilter = new List<string>();

        Node clip = null;

        private int left;
        private int hidden;

        public AddItem addItemForm;
        public PrefsEditor prefsForm;
        public Prefs prefs;

        public void Setup()
        {
            inhibitAutoSave = true;

            ResetList();

            prefs = new Prefs();
            prefs.list = this;
            prefs.LoadPrefs();

            if ((prefs.loadLastList == false) || (prefs.lastFile == null) ||
                prefs.lastFile.Equals("") || (File.Exists(prefs.lastFile) == false))
            {
                NewList();
            }
            else
            {
                LoadFile(prefs.lastFile);
            }

            SetListFont();

            inhibitAutoSave = false;
        }

        public new void Dispose()
        {
            visibleNode = null;
            rootNode = null;

            base.Dispose();
        }

        // general document handling

        public void ResetList()
        {
            rootNode = new Node("root", this);
            visibleNode = rootNode;

            treeView.Nodes.Clear();

            fileName = null;
            modified = false;

            title = "New list";

            undoFrames = new Stack<Stack<UndoFrame>>();
        }

        private void NewList()
        {
            ResetList();

            UpdateTreeView(null);

            //titleBox.Focus();
            //titleBox.SelectAll();
        }

        private void LoadFile(String fn)
        {
            Dictionary<int, Node> lastNodes = new Dictionary<int, Node>();

            ResetList();

            inhibitAutoSave = true;

            fileName = fn;
            title = "";

            try
            {
                using (StreamReader sr = new StreamReader(fn))
                {
                    String line;

                    while ((line = sr.ReadLine()) != null)
                    {
                        int indent = 0;

                        while (line.StartsWith("  "))
                        {
                            indent++;
                            line = line.Substring(2);
                        }

                        line = line.Trim();

                        if (line.Length == 0)
                            continue;

                        // first non blank line is the title

                        if (title == "")
                        {
                            title = line;
                            continue;
                        }

                        bool selected = false;
                        bool open = false;

                        if (line.StartsWith("*"))
                        {
                            selected = true;
                            line = line.Substring(1);
                        }

                        if (line.StartsWith("+"))
                        {
                            open = true;
                            line = line.Substring(1);
                        }

                        // parse tags

                        int openSquareBracket = line.IndexOf("[");
                        string tagText = null;

                        if (openSquareBracket != -1)
                        {
                            int closeSquareBracket = line.IndexOf("]");

                            if ((closeSquareBracket != -1) && (closeSquareBracket > openSquareBracket))
                            {
                                tagText = line.Substring(openSquareBracket + 1,
                                    closeSquareBracket - openSquareBracket - 1);

                                line = line.Remove(openSquareBracket, closeSquareBracket - openSquareBracket + 1);
                            }
                        }

                        Node item = new Node(line.Trim(), this);

                        item.selected = selected;
                        item.open = open;

                        if (tagText != null)
                        {
                            string[] tagNames = tagText.Split(',');

                            foreach (String tagName in tagNames)
                            {
                                item.AddTag(tagName.Trim());
                            }
                        }

                        if (indent == 0)
                        {
                            rootNode.Add(item);

                            lastNodes[indent] = item;
                        }
                        else
                        {
                            // Insert "Unnamed" nodes if list is missing them

                            while (lastNodes.Count < indent)
                            {
                                int newIndent = lastNodes.Count;

                                Node unnamedNode = new Node("Unnamed", this);

                                if (newIndent == 0)
                                {
                                    rootNode.Add(unnamedNode);
                                }
                                else
                                {
                                    lastNodes[newIndent - 1].Add(unnamedNode);
                                }

                                lastNodes[newIndent] = unnamedNode;
                            }

                            Node group = lastNodes[indent - 1];

                            if (group != null)
                            {
                                group.Add(item);

                                lastNodes[indent] = item;
                            }
                        }
                    }
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
            }

            inhibitAutoSave = false;

            CommonShowAll(null, null);

            if (treeView.Nodes.Count > 0)
                treeView.SelectedNode = treeView.Nodes[0];

            treeView.Focus();

            modified = false;
        }

        public void ReportSave(bool saveAs)
        {
            string lastPart = fileName.Substring(fileName.LastIndexOf("\\") + 1);
            string firstPart = "Saved to ";

            if (saveAs)
            {
                firstPart = "Copy saved to ";
            }

            MessageBox.Show(firstPart + lastPart, "Save successful");
        }

        private void SaveFileAux(List<Node> nodes, int indent, StreamWriter sw)
        {
            foreach (Node node in nodes)
            {
                string line = "";

                for (int i = 0; i < indent; i++)
                    line += "  ";

                if (node.items.Count == 0)
                {
                    if (node.selected)
                        line += "*";
                }
                else
                {
                    if (node.open)
                        line += "+";
                }

                line += node.text;

                if (node.tags.Count > 0)
                {
                    line += " [";

                    for (int i = 0; i < node.tags.Count; i++)
                    {
                        line += node.tags[i];

                        if (i < (node.tags.Count - 1))
                        {
                            line += ",";
                        }
                    }

                    line += "]";
                }

                sw.WriteLine(line);

                if (node.items.Count > 0)
                    SaveFileAux(node.items, indent + 1, sw);
            }
        }

        private void SaveFile(string fn)
        {
            try
            {
                using (StreamWriter sw = new StreamWriter(fn))
                {
                    sw.WriteLine(title);

                    SaveFileAux(rootNode.items, 0, sw);

                    modified = false;
                }
            }
            catch (Exception e)
            {
                MessageBox.Show(e.Message);
            }
        }

        public void AutoSave()
        {
            if (inhibitAutoSave == false)
            {
                if (prefs.autoSave)
                {
                    if (fileName != null)
                    {
                        SaveFile(fileName);
                    }
                }
            }
        }

        public void setVisibleRoot(Node node)
        {
            visibleNode = node;

            UpdateTitleBox();
            UpdateTreeView(null);
        }

        public void ListChanged()
        {
            modified = true;
            Sanitise();

            UpdateTreeView(null);
            AutoSave();
        }

        // node searching

        private TreeNode FindTreeNodeAux(TreeNodeCollection treeNodes, Node searchNode)
        {
            for (int i = 0; i < treeNodes.Count; i++)
            {
                TreeNode treeNode = treeNodes[i];
                Node node = (Node)treeNode.Tag;

                if (node == searchNode)
                    return treeNode;

                TreeNode result = FindTreeNodeAux(treeNode.Nodes, searchNode);

                if (result != null)
                    return result;
            }

            return null;
        }

        public TreeNode FindTreeNode(Node node)
        {
            return FindTreeNodeAux(treeView.Nodes, node);
        }

        private bool FindAboveNodes(List<Node> results, Node node, Node searchNode, int depth)
        {
            if (node == searchNode)
                return true;

            while (results.Count <= depth)
                results.Add(null);

            results[depth] = node;

            foreach (Node n in node.items)
            {
                if (FindAboveNodes(results, n, searchNode, depth + 1))
                {
                    return true;
                }
            }

            return false;
        }

        public List<int> calculateContext(Node node)
        {
            List<int> context = new List<int>();

            while (node.parent != null)
            {
                context.Insert(0, node.parent.items.IndexOf(node));

                node = node.parent;
            }

            return context;
        }

        public Node getContext(List<int> context)
        {
            Node result = rootNode;

            foreach (int x in context)
            {
                if (x >= result.items.Count)
                    return null;

                result = result.items[x];
            }

            return result;
        }

        public List<String> allTags
        {
            get
            {
                List<String> result = rootNode.cachedTags;
                result.Sort();

                return result;
            }
        }

        // update routines

        public void UpdateTitleBox()
        {
            if (visibleNode == rootNode)
            {
                titleBox.Text = _title;
                Text = _title;
            }
            else
            {
                titleBox.Text = visibleNode.text;
            }
        }

        private void UpdateTreeNodeState(TreeNode treeNode)
        {
            Node node = (Node)treeNode.Tag;

            bool currentCheck = treeNode.Checked;
            bool checkRequired = node.AllSelected();

            Color currentColor = treeNode.ForeColor;
            Color colorRequired = checkRequired ? checkedColor : uncheckedColor;

            string currentText = treeNode.Text;
            string textRequired = node.text;

            if (currentCheck != checkRequired)
                treeNode.Checked = checkRequired;

            if (currentColor != colorRequired)
                treeNode.ForeColor = colorRequired;

            if (String.Compare(currentText, textRequired) != 0)
                treeNode.Text = textRequired;

            if (node.IsGroup())
            {
                if (node.open)
                {
                    if (treeNode.IsExpanded == false)
                    {
                        treeNode.Expand();
                    }
                }
                else
                {
                    if (treeNode.IsExpanded == true)
                    {
                        treeNode.Collapse();
                    }
                }
            }
        }

        private void UpdateSummaryAux(List<Node> nodes)
        {
            foreach (Node node in nodes)
            {
                if (node.IsGroup())
                {
                    UpdateSummaryAux(node.items);
                }
                else
                {
                    if (node.selected == false)
                    {
                        left++;

                        if (node.visible == false)
                        {
                            hidden++;
                        }
                    }
                }
            }
        }

        private void UpdateSummary()
        {
            left = 0;
            hidden = 0;

            UpdateSummaryAux(rootNode.items);

            string txt;

            if (rootNode.items.Count == 0)
            {
                txt = "";
            }
            else if (left == 0)
            {
                txt = "List complete!";
            }
            else
            {
                txt = left + " item" + (left == 1 ? "" : "s") + " left";

                if (hidden > 0)
                {
                    txt += " [" + hidden + " hidden]";
                }

            }

            if ((visibleNode != rootNode) && (left > 0))
            {
                txt += " - limit";
            }

            SpecificUI.setStatusText(this, txt);
        }

        private void UpdateTreeNodeOpenState(TreeNodeCollection tnc)
        {
            foreach (TreeNode tn in tnc)
            {
                Node ln = (Node)tn.Tag;

                if (ln.IsGroup())
                {
                    if (ln.open)
                    {
                        tn.Expand();
                        UpdateTreeNodeOpenState(tn.Nodes);
                    }
                }
            }
        }

        private void UpdateTreeViewAux(List<Node> nodes, TreeNodeCollection treeNodes)
        {
            int i;

            // remove treeNodes

            for (i = 0; i < treeNodes.Count; )
            {
                TreeNode deleteCandidate = treeNodes[i];
                Node owner = (Node)deleteCandidate.Tag;

                if (nodes.IndexOf(owner) == -1)
                {
                    treeNodes.Remove(deleteCandidate);
                    continue;
                }

                if (owner.visible == false)
                {
                    treeNodes.Remove(deleteCandidate);
                    continue;
                }

                i++;
            }

            // add treeNodes

            int ti = 0;

            for (i = 0; i < nodes.Count; i++)
            {
                Node node = nodes[i];

                if (node.visible)
                {
                    TreeNode visibleTreeNode = ti < treeNodes.Count ? treeNodes[ti] : null;
                    Node visibleNode = ti < treeNodes.Count ? (Node)treeNodes[ti].Tag : null;

                    TreeNode treeNode;

                    if (visibleNode != node)
                    {
                        string text = node.text;

                        treeNode = new TreeNode(text);
                        treeNode.Tag = node;

                        UpdateTreeViewAux(node.items, treeNode.Nodes);
                        UpdateTreeNodeState(treeNode);

                        treeNodes.Insert(ti, treeNode);
                    }
                    else
                    {
                        UpdateTreeViewAux(node.items, visibleTreeNode.Nodes);
                        UpdateTreeNodeState(visibleTreeNode);
                    }

                    ti++;
                }
            }
        }

        public void UpdateTreeView(Node nodeToSelect)
        {
            //TreeNode topNode = treeView.TopNode;
            TreeNode selectedNode = treeView.SelectedNode;

            updateInProgress = true;
            treeView.BeginUpdate();

            Node root = visibleNode;

            UpdateTreeViewAux(root.items, treeView.Nodes);
            UpdateTreeNodeOpenState(treeView.Nodes);
            UpdateSummary();

            // Show the root lines?

            bool showRootLines = false;

            foreach (Node node in root.items)
            {
                if (node.IsGroup())
                {
                    showRootLines = true;
                }
            }

            treeView.ShowRootLines = showRootLines;

            //treeView.TopNode = topNode;

            if (nodeToSelect != null)
            {
                treeView.SelectedNode = FindTreeNode(nodeToSelect);
            }
            else
            {
                treeView.SelectedNode = selectedNode;
            }

            treeView.EndUpdate();
            updateInProgress = false;
        }

        public void SelectTreeNode(TreeNode treeNode)
        {
            treeView.SelectedNode = treeNode;
        }

        public void SelectNode(Node node)
        {
            treeView.SelectedNode = FindTreeNode(node);
        }

        // 

        public void TitleBoxChanged(object sender, EventArgs e)
        {
            if (visibleNode == rootNode)
            {
                _title = titleBox.Text;
                Text = titleBox.Text;
            }
            else
            {
                visibleNode.text = titleBox.Text;
            }

            AutoSave();
        }

        private void ShowAllAux(List<Node> nodes)
        {
            foreach (Node node in nodes)
            {
                node.visible = true;

                if (node.IsGroup())
                {
                    ShowAllAux(node.items);
                }

            }
        }

        private bool ShowTickedAux(List<Node> nodes)
        {
            bool show = false;

            foreach (Node node in nodes)
            {
                if (!node.IsGroup())
                {
                    node.visible = node.selected;
                }
                else
                {
                    node.visible = ShowTickedAux(node.items);
                }

                if (node.visible)
                    show = true;
            }

            return show;
        }

        private bool ShowUntickedAux(List<Node> nodes)
        {
            bool show = false;

            foreach (Node node in nodes)
            {
                if (!node.IsGroup())
                {
                    node.visible = !node.selected;
                }
                else
                {
                    node.visible = ShowUntickedAux(node.items);
                }

                if (node.visible)
                    show = true;
            }

            return show;
        }

        private void UncheckAllAux(List<Node> nodes)
        {
            foreach (Node node in nodes)
            {
                if (!node.IsGroup())
                {
                    node.selected = false;
                }
                else
                {
                    UncheckAllAux(node.items);
                }

                node.visible = true;
            }
        }

        private void CheckAllAux(List<Node> nodes)
        {
            foreach (Node node in nodes)
            {
                if (!node.IsGroup())
                {
                    node.selected = true;
                }
                else
                {
                    CheckAllAux(node.items);
                }
            }
        }

        private void CommonShowAll(object sender, EventArgs e)
        {
            visibleNode = rootNode;
            tagFilter.Clear();

            UpdateTitleBox();

            ShowAllAux(rootNode.items);
            UpdateTreeView(null);
        }

        private void CommonShowTicked(object sender, EventArgs e)
        {
            ShowTickedAux(rootNode.items);
            UpdateTreeView(null);
        }

        private void CommonShowUnticked(object sender, EventArgs e)
        {
            ShowUntickedAux(rootNode.items);
            UpdateTreeView(null);
        }

        private bool RevealUncheckedAux(Node node)
        {
            node.visible = true;

            if (node.IsGroup())
            {
                bool reveal = false;

                foreach (Node subNode in node.items)
                {
                    if (RevealUncheckedAux(subNode))
                    {
                        reveal = true;
                    }
                }

                node.open = reveal;
                return reveal;
            }
            else
            {
                return node.selected == false;
            }
        }

        private void CommonRevealUnchecked(object sender, EventArgs e)
        {
            RevealUncheckedAux(visibleNode);
            UpdateTreeView(null);
            UpdateTitleBox();
        }

        private void CommonUncheckAll(object sender, EventArgs e)
        {
            DialogResult result = MessageBox.Show("Are you sure want to uncheck all items?",
                "Comfirm uncheck all", MessageBoxButtons.YesNo, MessageBoxIcon.Question,
                MessageBoxDefaultButton.Button1);

            if (result.ToString() == "No")
                return;

            AddUndoFrame(this, UndoFrame.MODE_MODIFY, rootNode);

            UncheckAllAux(rootNode.items);
            modified = true;
            UpdateTreeView(null);

            AutoSave();
        }

        private void CommonExit(object sender, EventArgs e)
        {
            if (modified == true)
            {
                DialogResult result = MessageBox.Show("You have unsaved changes.  Are you sure you want" +
                    " to exit?",
                    "Comfirm exit", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation,
                    MessageBoxDefaultButton.Button1);

                if (result.ToString() == "No")
                    return;
            }

            prefs.SavePrefs();
            System.Windows.Forms.Application.Exit();
        }

        private void CommonItemChecked(object sender, TreeViewEventArgs e)
        {
            if (updateInProgress)
                return;

            modified = true;

            Node item = (Node)e.Node.Tag;

            if (!((Node)e.Node.Tag).IsGroup())
            {
                AddUndoFrame(this, UndoFrame.MODE_MODIFY, item);
                item.selected = e.Node.Checked;
                ListChanged();
            }
            else
            {
                string word = "check";

                if (e.Node.Checked == false)
                    word = "uncheck";

                DialogResult result = MessageBox.Show("Are you sure want to " + word + " \"" +
                    item.text + "\"?",
                    "Comfirm group action", MessageBoxButtons.YesNo, MessageBoxIcon.Question,
                    MessageBoxDefaultButton.Button1);

                if (result.ToString() == "No")
                {
                    UpdateTreeView(null);
                    return;
                }

                // toggle the group

                AddUndoFrame(this, UndoFrame.MODE_MODIFY, item);

                if (e.Node.Checked)
                {
                    CheckAllAux(item.items);
                }
                else
                {
                    UncheckAllAux(item.items);
                }

                ListChanged();
            }

            UpdateSummary();
        }

        private void CommonDeleteItem(Node item)
        {
            modified = true;

            Node parent = item.parent;
            int index = parent.items.IndexOf(item);

            AddUndoFrame(this, UndoFrame.MODE_REMOVE, item);
            parent.RemoveAt(index);
            ListChanged();
        }

        private void DeleteItem(object sender, EventArgs e)
        {
            Node item = (Node)treeView.SelectedNode.Tag;

            DialogResult result = MessageBox.Show("Are you sure want to delete \"" +
                item.text + "\"?",
                "Comfirm delete", MessageBoxButtons.YesNo, MessageBoxIcon.Exclamation,
                MessageBoxDefaultButton.Button1);

            if (result.ToString() == "No")
                return;

            CommonDeleteItem(item);
        }

        private void CommonGroupClosed(object sender, TreeViewEventArgs e)
        {
            ((Node)e.Node.Tag).open = false;
        }

        private void CommonGroupOpened(object sender, TreeViewEventArgs e)
        {
            ((Node)e.Node.Tag).open = true;
        }

        private void CommonAdd(object sender, EventArgs e)
        {
            addItemForm = new AddItem(this, null, "Add item", sender, "Add item to list:", AddItem.MODE_APPEND);

            addItemForm.ShowDialog();
        }

        private void InsertBefore(object sender, EventArgs e)
        {
            Node item = (Node)treeView.SelectedNode.Tag;

            addItemForm = new AddItem(this, null, "Insert item", treeView.SelectedNode, "Insert before \"" +
                item.text + "\":", AddItem.MODE_INSERT_BEFORE);

            addItemForm.ShowDialog();
        }

        private void InsertAfter(object sender, EventArgs e)
        {
            Node item = (Node)treeView.SelectedNode.Tag;

            addItemForm = new AddItem(this, null, "Insert item", treeView.SelectedNode, "Insert after \"" +
                item.text + "\":", AddItem.MODE_INSERT_AFTER);

            addItemForm.ShowDialog();
        }

        private void CreateSubItemClicked(object sender, EventArgs e)
        {
            addItemForm = new AddItem(this, null, "Create sub item", treeView.SelectedNode, "Create sub item under \"" +
                ((Node)treeView.SelectedNode.Tag).text + "\":",
                AddItem.MODE_CREATE_SUB);

            addItemForm.ShowDialog();
        }

        private void CommonSaveClicked(object sender, EventArgs e)
        {
            if (fileName == null)
            {
                DialogResult dr = saveFileDialog.ShowDialog();

                if (dr.ToString() != "OK")
                {
                    return;
                }

                fileName = saveFileDialog.FileName;
            }

            SaveFile(fileName);
            ReportSave(false);
        }

        private void CommonSaveAsClicked(object sender, EventArgs e)
        {
            DialogResult dr = saveFileDialog.ShowDialog();

            if (dr.ToString() == "OK")
            {
                fileName = saveFileDialog.FileName;

                SaveFile(fileName);
                ReportSave(true);
            }
        }

        private void CommonCut()
        {
            if (treeView.SelectedNode == null)
                return;

            clip = ((Node)treeView.SelectedNode.Tag);

            AddUndoFrame(this, UndoFrame.MODE_REMOVE, clip);
            clip.parent.Remove(clip);

            ListChanged();
        }

        private void CommonCopy()
        {
            if (treeView.SelectedNode == null)
                return;

            clip = ((Node)treeView.SelectedNode.Tag).Copy();
        }

        private void CommonPaste()
        {
            PasteAux(false);
        }

        private void CutClicked(object sender, EventArgs e)
        {
            CommonCut();
        }

        private void CopyClicked(object sender, EventArgs e)
        {
            CommonCopy();
        }

        private void PasteAux(bool after)
        {
            if (clip == null)
                return;

            List<Node> target;
            Node targetLN;
            int index;

            if (treeView.SelectedNode == null)
            {
                target = visibleNode.items;
                targetLN = visibleNode;
                index = 0;
            }
            else
            {
                Node thisNode = (Node)treeView.SelectedNode.Tag;
                Node parentNode = thisNode.parent;

                target = parentNode.items;
                targetLN = parentNode;

                index = target.IndexOf(thisNode);
            }

            Node copy = clip.Copy();
            copy.parent = targetLN;

            if (after)
                index++;

            target.Insert(index, copy);
            AddUndoFrame(this, UndoFrame.MODE_INSERT, copy);
            ListChanged();
        }

        private void editPasteMenuItem_Click(object sender, EventArgs e)
        {
            PasteAux(false);
        }

        private void PasteBeforeClicked(object sender, EventArgs e)
        {
            PasteAux(false);
        }

        private void PasteAfterClicked(object sender, EventArgs e)
        {
            PasteAux(true);
        }

        private void PasteIntoClicked(object sender, EventArgs e)
        {
            if (treeView.SelectedNode == null)
            {
                PasteAux(true);
                return;
            }

            Node current = (Node)treeView.SelectedNode.Tag;
            Node copy = clip.Copy();

            current.Add(copy);
            current.open = true;

            AddUndoFrame(this, UndoFrame.MODE_INSERT, copy);
            ListChanged();
        }

        private void EditClicked(object sender, EventArgs e)
        {
            addItemForm = new AddItem(this, ((Node)treeView.SelectedNode.Tag), "Edit item", treeView.SelectedNode, "Edit \"" +
                ((Node)treeView.SelectedNode.Tag).text + "\":",
                AddItem.MODE_EDIT);

            addItemForm.ShowDialog();
        }

        private void ExpandCollapseItemAux(Node node, bool newState)
        {
            node.open = newState;

            foreach (Node child in node.items)
            {
                ExpandCollapseItemAux(child, newState);
            }
        }

        private void ExpandItem(object sender, EventArgs e)
        {
            ExpandCollapseItemAux((Node)(treeView.SelectedNode.Tag), true);
            UpdateTreeView(null);
        }

        private void CollapseItem(object sender, EventArgs e)
        {
            ExpandCollapseItemAux((Node)(treeView.SelectedNode.Tag), false);
            UpdateTreeView(null);
        }

        private void CommonContextPopup(object sender, EventArgs e)
        {
            bool treeVisible = false;
            string nodeName = "";

            if (treeView.SelectedNode != null)
            {
                Node node = (Node)treeView.SelectedNode.Tag;

                treeVisible = node.items.Count > 0;

                nodeName = node.text;
            }

            SpecificUI.Clear(this);

            bool cutCopyEnabled = treeView.SelectedNode != null;

            SpecificUI.Add(this, "Cut", false, cutCopyEnabled, new EventHandler(CutClicked));
            SpecificUI.Add(this, "Copy", false, cutCopyEnabled, new EventHandler(CopyClicked));

            if (treeView.SelectedNode != null)
            {
                SpecificUI.Add(this, "Paste before", false, clip != null, new EventHandler(PasteBeforeClicked));
                SpecificUI.Add(this, "Paste after", false, clip != null, new EventHandler(PasteAfterClicked));
                SpecificUI.Add(this, "Paste into", false, clip != null, new EventHandler(PasteIntoClicked));
            }
            else
            {
                SpecificUI.Add(this, "Paste", false, clip != null, new EventHandler(PasteBeforeClicked));
            }

            SpecificUI.AddSeparator(this, false);

            SpecificUI.Add(this, "Insert before...", false, treeView.SelectedNode != null, new EventHandler(InsertBefore));
            SpecificUI.Add(this, "Insert after...", false, treeView.SelectedNode != null, new EventHandler(InsertAfter));
            SpecificUI.Add(this, "Create sub item...", false, treeView.SelectedNode != null, new EventHandler(CreateSubItemClicked));

            SpecificUI.AddSeparator(this, false);

            if (treeVisible)
            {
                SpecificUI.Add(this, "Tree", false, true, null);

                SpecificUI.Add(this, "Expand all \"" + nodeName + "\"", true, true, new EventHandler(ExpandItem));
                SpecificUI.Add(this, "Collapse all \"" + nodeName + "\"", true, true, new EventHandler(CollapseItem));
                SpecificUI.AddSeparator(this, true);
                SpecificUI.Add(this, "Limit view", true, true, new EventHandler(LimitView));

                SpecificUI.AddSeparator(this, false);
            }

            SpecificUI.Add(this, "Edit...", false, treeView.SelectedNode != null, new EventHandler(EditClicked));
            SpecificUI.Add(this, "Delete...", false, treeView.SelectedNode != null, new EventHandler(DeleteItem));
        }

        private void CommonExpandAllClicked(object sender, EventArgs e)
        {
            foreach (Node node in rootNode.items)
            {
                ExpandCollapseItemAux(node, true);
            }

            UpdateTreeView(null);
        }

        private void CommonCollapseAllClicked(object sender, EventArgs e)
        {
            foreach (Node node in rootNode.items)
            {
                ExpandCollapseItemAux(node, false);
            }

            UpdateTreeView(null);
        }

        private void CommonPreferencesClicked(object sender, EventArgs e)
        {
            prefsForm = new PrefsEditor(prefs);
        }

        private void ScopeIn()
        {
            if (treeView.SelectedNode == null)
                return;

            setVisibleRoot((Node)treeView.SelectedNode.Tag);
        }

        private void ScopeOut()
        {
            if (visibleNode == rootNode)
                return;

            setVisibleRoot(visibleNode.parent);
        }

        private void LimitView(object sender, EventArgs e)
        {
            if (treeView.SelectedNode == null)
                return;

            setVisibleRoot((Node)treeView.SelectedNode.Tag);
        }

        private void CommonAbout(object sender, EventArgs e)
        {
            MessageBox.Show("List for Windows.\n\nDon Cruickshank.", "About List...");
        }

        private string PrintNodeList(List<Node> nodes)
        {
            string result = "[";

            for (int i = 0; i < nodes.Count; i++)
            {
                result += nodes[i].text;

                if (i < (nodes.Count - 1))
                    result += ", ";
            }

            result += "]";

            return result;
        }

        // Undo code

        public Stack<Stack<UndoFrame>> undoFrames = new Stack<Stack<UndoFrame>>();

        public UndoFrame CreateUndoFrame(List list, int type, Node node)
        {
            if ((type == UndoFrame.MODE_MODIFY) && (node == rootNode))
                type = UndoFrame.MODE_MODIFY_ROOT;

            Node context = node.parent;
            int index = 0;

            if (context != null)
                index = context.items.IndexOf(node);

            return new UndoFrame(list, type, index, node.Copy(), context);
        }

        public void AddUndoFrame(List list, int type, Node node)
        {
            Stack<UndoFrame> frames = new Stack<UndoFrame>();

            frames.Push(CreateUndoFrame(list, type, node));

            AddUndoFrames(frames);
        }

        public void AddUndoFrames(Stack<UndoFrame> frames)
        {
            undoFrames.Push(frames);
        }

        private void CommonUndo()
        {
            if (undoFrames.Count == 0)
                return;

            Stack<UndoFrame> frames = undoFrames.Pop();

            while (frames.Count > 0)
            {
                UndoFrame uf = frames.Pop();

                //MessageBox.Show("About to modify: \"" + getUndoContext(uf).text + "\", index = " + uf.index);

                if (uf.type == UndoFrame.MODE_INSERT)
                {
                    getContext(uf.context).RemoveAt(uf.index);
                }
                else if (uf.type == UndoFrame.MODE_MODIFY)
                {
                    getContext(uf.context).ReplaceAt(uf.index, uf.node);
                }
                else if (uf.type == UndoFrame.MODE_REMOVE)
                {
                    getContext(uf.context).Insert(uf.index, uf.node);
                }
                else if (uf.type == UndoFrame.MODE_MODIFY_ROOT)
                {
                    rootNode = uf.node;
                    visibleNode = uf.node;
                }
            }

            ListChanged();
        }

        public void CommonEditMenuOpening()
        {
            undoMenuItem.Enabled = undoFrames.Count > 0;
        }

        private void Sanitise()
        {
            // check tag filter and remove non existent tags

            List<String> tags = allTags;

            foreach (String tag in tagFilter.ToArray())
            {
                if (tags.Contains(tag) == false)
                {
                    tagFilter.Remove(tag);
                }
            }
        }

        public void SetListFont()
        {
            if ((prefs.fontFamily.Length > 0) && (prefs.fontSize > 0))
            {
                treeView.Font = new Font(prefs.fontFamily, prefs.fontSize, prefs.fontStyle);
            }

            if (prefs.fontSize > 0)
            {
                if (fontSizes != null)
                {
                    int sliderPosition = 0;

                    for (int i = 0; i < fontSizes.Count; i++)
                    {
                        if (prefs.fontSize >= (int)fontSizes[i])
                        {
                            sliderPosition = i;
                        }
                    }

                    zoomBar.Value = sliderPosition;
                }
            }
        }

        // List filters

        private bool UpdateTagFilterVisibleStateAux(List<Node> nodes)
        {
            bool show = false;

            foreach (Node node in nodes)
            {
                node.visible = false;

                if (!node.IsGroup())
                {
                    if (tagFilter.Count == 0)
                    {
                        node.visible = true;
                    }
                    else
                    {
                        foreach (String tag in node.tags)
                        {
                            if (tagFilter.Contains(tag))
                            {
                                node.visible = true;
                                break;
                            }
                        }
                    }
                }
                else
                {
                    foreach (String tag in node.tags)
                    {
                        if (tagFilter.Contains(tag))
                        {
                            node.visible = true;
                            break;
                        }
                    }

                    if (node.visible == true)
                    {
                        ShowAllAux(node.items);
                    }
                    else
                    {
                        node.visible = UpdateTagFilterVisibleStateAux(node.items);
                    }
                }

                if (node.visible)
                {
                    show = true;
                }
            }

            return show;
        }

        private void UpdateTagFilterVisibleState()
        {
            Sanitise();

            UpdateTagFilterVisibleStateAux(rootNode.items);
            UpdateTreeView(null);
        }
    }

    public class Node
    {
        public List<Node> items = new List<Node>();
        public Node parent = null;
        public string text;
        public Boolean selected = false;
        public bool visible = true;
        public bool open = false;
        public List list;

        public Node(string newText, List theList)
        {
            text = newText;
            list = theList;
        }

        public bool AllSelected()
        {
            if (items.Count == 0)
            {
                return selected;
            }
            else
            {
                foreach (Node node in items)
                {
                    if (node.AllSelected() == false)
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        public bool IsGroup()
        {
            return (items.Count > 0);
        }

        public Node Copy()
        {
            Node node = new Node(text, list);

            node.open = open;
            node.selected = selected;

            foreach (Node subNode in items)
            {
                node.Add(subNode.Copy());
            }

            return node;
        }

        public void SetList(List theList)
        {
            list = theList;

            foreach (Node subNode in items)
            {
                subNode.SetList(theList);
            }
        }

        public void Add(Node node)
        {
            if (node.parent != null)
                node.parent.items.Remove(node);

            node.parent = this;
            items.Add(node);

            node.UpdateTagCache();
        }

        public void Insert(int index, Node node)
        {
            if (node.parent != null)
                node.parent.items.Remove(node);

            node.parent = this;
            items.Insert(index, node);

            node.UpdateTagCache();
        }

        public void ReplaceAt(int index, Node node)
        {
            if (node.parent != null)
                node.parent.items.Remove(node);

            node.parent = this;
            items[index] = node;

            node.UpdateTagCache();
        }

        public void Remove(Node node)
        {
            Node parent = node.parent;

            node.parent = null;
            items.Remove(node);

            parent.UpdateTagCache();
        }

        public void RemoveAt(int index)
        {
            Node parent = items[index].parent;

            items[index].parent = null;
            items.RemoveAt(index);

            parent.UpdateTagCache();
        }

        public int IndexOf(Node node)
        {
            return items.IndexOf(node);
        }

        public int Count
        {
            get
            {
                return items.Count;
            }
        }

        public int VisibleCount
        {
            get
            {
                int result = 0;

                if (open)
                {

                    foreach (Node node in items)
                    {
                        if (node.visible)
                        {
                            result++;
                        }
                    }
                }

                return result;
            }
        }

        public Node PrevSibling()
        {
            if (parent == null)
                return null;

            int index = parent.items.IndexOf(this);

            if (index == 0)
                return null;

            return parent.items[index - 1];
        }

        public Node NextSibling()
        {
            if (parent == null)
                return null;

            int index = parent.items.IndexOf(this);

            if (index == (parent.items.Count - 1))
                return null;

            return parent.items[index + 1];
        }

        // Tag support

        public List<String> tags = new List<String>();
        public List<String> cachedTags = new List<String>();

        public void UpdateTagCache()
        {
            cachedTags.Clear();
            cachedTags.AddRange(tags);

            foreach (Node node in items)
            {
                foreach (String tag in node.cachedTags)
                {
                    if (cachedTags.Contains(tag) == false)
                        cachedTags.Add(tag);
                }
            }

            if (parent != null)
            {
                parent.UpdateTagCache();
            }
        }

        public void AddTag(string tag)
        {
            if (tags.Contains(tag))
                return;

            tags.Add(tag);
            UpdateTagCache();
        }
    }

    public class UndoFrame
    {
        public List<int> context = new List<int>();

        public static int MODE_INSERT = 0;
        public static int MODE_MODIFY = 1;
        public static int MODE_REMOVE = 2;
        public static int MODE_MODIFY_ROOT = 3;

        public int type;
        public int index;
        public Node node;

        public string debugString()
        {
            string str = "[";

            for (int i = 0; i < context.Count; i++)
            {
                str += context[i];
                if (i < (context.Count - 1))
                    str += ",";
            }

            return str;
        }

        public UndoFrame(List list, int theType, int theIndex, Node theNode, Node referenceNode)
        {
            type = theType;
            index = theIndex;
            node = theNode;

            if (referenceNode != null)
            {
                context = list.calculateContext(referenceNode);
            }
        }
    }

    public partial class AddItem : Form
    {
        public Object context;
        public int mode;
        public List list;

        public const int MODE_APPEND = 0;
        public const int MODE_INSERT_BEFORE = 1;
        public const int MODE_INSERT_AFTER = 2;
        public const int MODE_CREATE_SUB = 3;
        public const int MODE_EDIT = 4;

        public List<String> availableTags = new List<string>();
        public List<String> tags = new List<string>();

        public AddItem(List theList, Node theNode, String thisText, Object thisContext, string thisLabel, int thisMode)
        {
            InitializeComponent();

            list = theList;
            Text = thisText;

            if (thisMode == MODE_EDIT)
            {
                Node original = (Node)((TreeNode)thisContext).Tag;
                newItem.Text = original.text;
            }

            availableTags = new List<string>();
            availableTags.AddRange(list.allTags);

            context = thisContext;
            SetLabelText(thisLabel);
            mode = thisMode;

            if (theNode != null)
            {
                tags.AddRange(theNode.tags);
            }
            else
            {
                tags.AddRange(list.tagFilter);
            }

            UpdateTagList();
        }

        private bool tagViewUpdateInProgress = false;

        private void UpdateTagList()
        {
            tagViewUpdateInProgress = true;

            availableTags.Sort();
            tagsTreeView.Nodes.Clear();

            foreach (String tag in availableTags)
            {
                TreeNode treeNode = new TreeNode(tag);

                if (tags.Contains(tag))
                {
                    treeNode.Checked = true;
                }

                tagsTreeView.Nodes.Add(treeNode);
            }

            tagViewUpdateInProgress = false;
        }

        private void SetTags()
        {
            tags.Clear();

            foreach (TreeNode treeNode in tagsTreeView.Nodes)
            {
                if (treeNode.Checked)
                {
                    tags.Add(treeNode.Text);
                }
            }
        }

        private void UpdateTags(Node node)
        {
            node.tags.Clear();
            node.tags.AddRange(tags);
            node.UpdateTagCache();
        }

        private void tagsTreeView_AfterCheck(object sender, TreeViewEventArgs e)
        {
            if (tagViewUpdateInProgress == false)
            {
                SetTags();
            }
        }

        private void OKAddItem(object sender, EventArgs e)
        {
            newItem.Text = newItem.Text.Trim();

            list.modified = true;

            if ((mode == MODE_APPEND) || (mode == MODE_INSERT_BEFORE) ||
                (mode == MODE_INSERT_AFTER) || (mode == MODE_CREATE_SUB))
            {
                if (list.prefs.autoCapitalise)
                {
                    string text = newItem.Text;

                    if (text.Length > 0)
                    {
                        text = text.Substring(0, 1).ToUpper() +
                            text.Substring(1);
                    }

                    newItem.Text = text;
                }
            }

            if (mode == MODE_APPEND)
            {
                Node node = list.rootNode;

                if (list.visibleNode != list.rootNode)
                    node = list.visibleNode;

                Node newNode = new Node(newItem.Text, list);

                UpdateTags(newNode);

                node.Add(newNode);
                node.open = true;

                list.AddUndoFrame(list, UndoFrame.MODE_INSERT, newNode);
                list.ListChanged();

                list.SelectTreeNode(list.FindTreeNode(newNode));
            }
            else if (mode == MODE_INSERT_BEFORE)
            {
                Node current = (Node)((TreeNode)context).Tag;

                Node newNode = new Node(newItem.Text, list);
                int newIndex = current.parent.items.IndexOf(current);

                UpdateTags(newNode);

                current.parent.Insert(newIndex, newNode);
                list.AddUndoFrame(list, UndoFrame.MODE_INSERT, newNode);
                list.ListChanged();

                list.SelectTreeNode(list.FindTreeNode(newNode));
            }
            else if (mode == MODE_INSERT_AFTER)
            {
                Node current = (Node)((TreeNode)context).Tag;

                Node newNode = new Node(newItem.Text, list);
                int newIndex = current.parent.items.IndexOf(current) + 1;

                UpdateTags(newNode);

                current.parent.Insert(newIndex, newNode);

                list.AddUndoFrame(list, UndoFrame.MODE_INSERT, newNode);
                list.ListChanged();

                list.SelectTreeNode(list.FindTreeNode(newNode));
            }
            else if (mode == MODE_CREATE_SUB)
            {
                Node current = (Node)((TreeNode)context).Tag;

                Node newNode = new Node(newItem.Text, list);

                UpdateTags(newNode);

                current.Add(newNode);
                current.open = true;
                list.AddUndoFrame(list, UndoFrame.MODE_INSERT, newNode);
                list.ListChanged();

                list.SelectTreeNode(list.FindTreeNode(newNode));
            }
            else if (mode == MODE_EDIT)
            {
                Node current = (Node)((TreeNode)context).Tag;

                list.AddUndoFrame(list, UndoFrame.MODE_MODIFY, current);
                current.text = newItem.Text;
                UpdateTags(current);
                list.ListChanged();
            }
        }

        public void Reset()
        {
            newItem.Text = "";
        }

        public string GetNewItemText()
        {
            return newItem.Text;
        }

        public void SetLabelText(string newText)
        {
            label1.Text = newText;
        }

        private void NewTag_Click(object sender, EventArgs e)
        {
            TagCreate form = new TagCreate();
            form.ShowDialog();

            if (form.DialogResult == DialogResult.OK)
            {
                if (availableTags.Contains(form.tagBox.Text) == false)
                {
                    availableTags.Add((string)form.tagBox.Text.Clone());
                    UpdateTagList();
                }
            }
        }
    }

    public class Prefs
    {
        public List list;

        public bool autoCapitalise;
        public bool loadLastList;
        public bool autoSave;
        public string lastFile = null;

        public string fontFamily;
        public float fontSize;
        public FontStyle fontStyle;

        private static string subKey = "Software\\Don Cruickshank\\list";

        private string hive = Registry.CurrentUser + "\\" + subKey;

        private const string autoCapitaliseHK = "Auto capitalise";
        private const string lastFileHK = "Last file";
        private const string loadLastListHK = "Load last file";
        private const string autoSaveHK = "Auto save";
        private const string fontFamilyHK = "Font family";
        private const string fontSizeHK = "Font size";
        private const string fontStyleHK = "Font style";

        private object getPreference(string key, object defaultValue)
        {
            object value = Registry.GetValue(hive, key, defaultValue);

            if (value == null)
                return defaultValue;

            return value;
        }

        public void LoadPrefs()
        {
            autoCapitalise = ((int)getPreference(autoCapitaliseHK, 0) == 1);
            loadLastList = ((int)getPreference(loadLastListHK, 1) == 1);
            autoSave = ((int)getPreference(autoSaveHK, 1) == 1);
            lastFile = (string)getPreference(lastFileHK, null);
            fontFamily = (string)getPreference(fontFamilyHK, "");
            fontSize = float.Parse((string)getPreference(fontSizeHK, "0"));

            int style = (int)getPreference(fontStyleHK, 0);

            fontStyle = FontStyle.Regular;

            if ((style & 1) == 1)
                fontStyle |= FontStyle.Bold;

            if ((style & 2) == 2)
                fontStyle |= FontStyle.Italic;

            if ((style & 4) == 4)
                fontStyle |= FontStyle.Underline;

            if ((style & 8) == 8)
                fontStyle |= FontStyle.Strikeout;
        }

        public void SavePrefs()
        {
            RegistryKey registryKey = Registry.CurrentUser.OpenSubKey(subKey, true);

            if (registryKey == null)
            {
                registryKey = Registry.CurrentUser.CreateSubKey(subKey);
            }

            if (registryKey == null)
            {
                MessageBox.Show("Unable to save preferences...");
                return;
            }

            registryKey.SetValue(autoCapitaliseHK, autoCapitalise ? 1 : 0);
            registryKey.SetValue(loadLastListHK, loadLastList ? 1 : 0);
            registryKey.SetValue(autoSaveHK, autoSave ? 1 : 0);
            registryKey.SetValue(fontFamilyHK, fontFamily == null ? "" : fontFamily);
            registryKey.SetValue(fontSizeHK, fontSize);
            registryKey.SetValue(fontStyleHK, (int)fontStyle);

            if (list.fileName != null)
                Registry.SetValue(hive, lastFileHK, list.fileName);

            registryKey.Close();
        }
    }

    public partial class PrefsEditor : Form
    {
        public Prefs prefs;
        public List list;

        private bool inhibitSave = false;

        public PrefsEditor(Prefs thePrefs)
        {
            InitializeComponent();

            prefs = thePrefs;

            inhibitSave = true;

            autoCapitaliseCheckBox.Checked = prefs.autoCapitalise;
            loadLastListCheckBox.Checked = prefs.loadLastList;
            autoSaveCheckBox.Checked = prefs.autoSave;

            inhibitSave = false;

            ShowDialog();
        }

        private void SaveThePrefs(object sender, EventArgs e)
        {
            if (inhibitSave == false)
            {
                prefs.autoCapitalise = autoCapitaliseCheckBox.Checked;
                prefs.loadLastList = loadLastListCheckBox.Checked;
                prefs.autoSave = autoSaveCheckBox.Checked;

                prefs.SavePrefs();
            }
        }
    }
}
